#include <stdlib.h>
#include <string.h>
//Fast lossless image compression and decompression based on QOI
//https://qoiformat.org/qoi-specification.pdf
//Data should be image data as sequence of RBGA bytes
//Probably will fail awefully to compress nonimage data
#define CACHESIZE 43
#define OP_RGB (char)0xFE
#define OP_RGBA (char)0xFF
#define OP_INDEX (char)0
#define OP_DIFF (char)0x40
#define OP_LUMA (char)0x80
#define OP_RUN (char)0xC0
#define OP_CODE (char)0xC0
#define OP_ARGS (char)0x3F
#define OP_LUMA_ARG (char)0x3F
#define OP_INDEX_ARG (char)0x3F
#define HASH(C, H) ((C.red * 3 + C.green * 5 + C.blue * 7 + C.alpha * 11) % H)
#define LRS(a, b) ((unsigned)a >> b)
#define LOCALHASH(C, H) (H + (LRS(C.red + 8, 3) ^ LRS(C.green + 8, 3) ^ \
                              LRS(C.blue + 8, 3)) %                     \
                                 (64 - H))
#define TUBITRANGE(a, b) ((char)(a - b) > -3 && (char)(a - b) < 2)
#define COLORRANGES(a, b) TUBITRANGE(a.red, b.red) &&         \
                              TUBITRANGE(a.green, b.green) && \
                              TUBITRANGE(a.blue, b.blue)
#define EQCOLOR(a, b) (a.red == b.red && a.green == b.green && \
                       a.blue == b.blue && a.alpha == b.alpha)

typedef struct RBGA
{
    unsigned char red;
    unsigned char green;
    unsigned char blue;
    unsigned char alpha;
} color;

int compress(const char *data, size_t len, char **result, size_t *outlen)
{
    color cache[64] = {0};
    color last;
    color current = (color){.alpha = 255};
    color temp, temp2;
    int i;
    char j, k, l;
    int run = 0;
    int ct = 0;
    int clen = CACHESIZE;
    if (len % 4)
    {
        return -1;
    }
    if (*result == NULL)
    {
        *result = (char *)malloc(5 * len);
    }
    for (i = 0; i < len; i += 4)
    {
        //Cache previous pixel
        cache[HASH(current, clen)] = current;
        last = current;

        //Get next pixel
        memcpy(&current, &data[i], 4);

        //Try to make run
        if (EQCOLOR(current, last) && run < 62)
        {
            run++;
            continue;
        }
        if (run)
        {
            (*result)[ct++] = OP_RUN | (run - 1);
            run = 0;
            if (EQCOLOR(current, last))
            {
                run++;
                continue;
            }
        }

        //Try to make exact diff with previous pixel
        if (COLORRANGES(current, last) &&
            current.alpha == last.alpha)
        {
            (*result)[ct++] = OP_DIFF | (current.red - last.red + 2 & 3) << 4 |
                              (current.green - last.green + 2 & 3) << 2 |
                              (current.blue - last.blue + 2 & 3);
            continue;
        }

        //Try to make exact index into cache
        temp = cache[HASH(current, clen)];
        if (EQCOLOR(current, temp))
        {
            (*result)[ct++] = OP_INDEX | HASH(current, clen) & OP_INDEX_ARG;
            continue;
        }

        //Try to make luma diff with previous pixel
        j = current.green - last.green;
        if (j > -33 && j < 32 && current.alpha == last.alpha)
        {
            k = current.red - last.red - j;
            l = current.blue - last.blue - j;
            if (-9 < k && -9 < l && k < 8 && l < 8)
            {
                (*result)[ct++] = OP_LUMA | (j + 32 & OP_LUMA_ARG);
                (*result)[ct++] = (k + 8 & 15) << 4 | l + 8 & 15;
                continue;
            }
        }
        if (64 - clen)
        {
            //Try to make diff index into cache
            j = LOCALHASH(current, clen);
            temp = cache[j];
            if (COLORRANGES(current, temp) &&
                current.alpha == temp.alpha)
            {
            smalldiff:
                (*result)[ct++] = OP_INDEX | j & OP_INDEX_ARG;
                (*result)[ct++] = OP_DIFF | (current.red - temp.red + 2 & 3) << 4 |
                                  (current.green - temp.green + 2 & 3) << 2 |
                                  (current.blue - temp.blue + 2 & 3);
                continue;
            }

            //Next just search the entire cache for the nearest color
            /*(This section removeable for a slight speed boost and a
            slight increase in compressed file size. I believe it
            only makes a difference when encoding very dark pixels
            nearish the beginning of the file.)*/
            for (j = clen; j < 64; j++)
            {
                temp2 = cache[j];
                if (COLORRANGES(current, temp2) && current.alpha == temp2.alpha)
                {
                    temp = temp2;
                    goto smalldiff;
                }
                k = current.green - temp2.green;
                if (k > -33 && k < 32)
                {
                    k = current.red - temp.red - j;
                    l = current.blue - temp.blue - j;
                    if (-9 < k && -9 < l && k < 8 && l < 8)
                    {
                        temp = temp2;
                        l = j;
                    }
                }
            }
            j = l;

            //Try to make luma index into cache
            k = current.green - temp.green;
            if (k > -33 && k < 32 && current.alpha == temp.alpha)
            {
                k = current.red - temp.red - j;
                l = current.blue - temp.blue - j;
                if (-9 < k && -9 < l && k < 8 && l < 8)
                {
                    (*result)[ct++] = OP_INDEX | j & OP_INDEX_ARG;
                    (*result)[ct++] = OP_LUMA | ((current.green - temp.green) + 32 & 0x3F);
                    (*result)[ct++] = (k + 8 & 15) << 4 | l + 8 & 15;
                    continue;
                }
            }
        }

        //Try to make RGB or RGBA pixel
        if (current.alpha == last.alpha)
        {
            (*result)[ct++] = OP_RGB;
            j = 3;
        }
        else
        {
            (*result)[ct++] = OP_RGBA;
            j = 4;
        }
        memcpy(&((*result)[ct]), &current, j);
        ct += j;
        if (clen - 64)
        {
            cache[LOCALHASH(current, clen)] = current;
        }
    }
    *outlen = (size_t)ct;
    return 0;
}

int decompress(const char *data, size_t len, char **result, size_t *outlen)
{
    color cache[64] = {0};
    color current = (color){.alpha = 255};
    int i = 0;
    char j;
    int buflen;
    char *tempbuf;
    int run = 0;
    int clen = CACHESIZE;

    if (*result == NULL)
    {
        buflen = 5 * len;
        *result = (char *)malloc(buflen);
    }
    else
    {
        buflen = *outlen & -4;
    }
    *outlen = 0;
    cache[HASH(current, clen)] = current;
    for (i = 0; i < len; i++)
    {
        //Expand buffer if it is full
        if (*outlen == buflen)
        {
            buflen *= 2;
            *result = (char *)realloc(*result, buflen);
        }

        //Add another pixel for current run
        if (run)
        {
            memcpy(&((*result)[*outlen]), &current, 4);
            *outlen += 4;
            run--;
            i--;
            continue;
        }

        //Decode next codeword
        switch (data[i] & OP_CODE)
        {

        case OP_INDEX:
            j = data[i] & OP_INDEX_ARG;
            current = cache[j];
            if (j < clen)
                break;
            i++;

        case OP_DIFF:
            if ((data[i] & OP_CODE) == OP_DIFF)
            {
                current.red += (LRS(data[i], 4) & 3) - 2;
                current.green += (LRS(data[i], 2) & 3) - 2;
                current.blue += (data[i] & 3) - 2;
                break;
            }

        case OP_LUMA:
            j = (data[i++] & OP_LUMA_ARG) - 32;
            current.green += j;
            current.red += j + (LRS(data[i], 4) & 0xF) - 8;
            current.blue += j + (data[i] & 0xF) - 8;
            break;

        case OP_RUN:
            if (data[i] == OP_RGB || data[i] == OP_RGBA)
            {
                memcpy(&current, &data[i + 1], 3);
                if (data[i] == OP_RGBA)
                {
                    current.alpha = data[i + 4];
                    i++;
                }
                i += 3;
                cache[LOCALHASH(current, clen)] = current;
            }
            else
            {
                run = (data[i] & OP_ARGS) + 1;
                continue;
            }
        }
        memcpy(&((*result)[*outlen]), &current, 4);
        cache[HASH(current, clen)] = current;
        *outlen += 4;
    }
    return 0;
}